/*
* This program is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.

* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.

* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

/*
* Programa:        SHELL_WRAPPER
* Funcao:          Disponibilizar opcoes de shell para  usuarios.  Apresenta
*                   um menu de operacoes disponiveis para o usuario em ques-
*                   tao, com relacao a um servidor especifico.
* Caracteristicas: Por padrao, para  todos os usuarios eh apresentada uma te-
*                   la contendo as opcoes disponiveis, mesmo que este usuario 
*                   possua apenas uma opcao neste servidor.
*                  Caso no final da descricao do shell, exista um '@', signi-
*                   fica que naquele servidor, o referido shell deve ser exe-
*                   cutado automaticamente, respeitando os seus argumentos.
* Versao:          1.2
* Changelog: 
*  1.2: Pequenos ajustes, alocacao dinamica, e melhor estruturacao. 
*  1.1: Suporte a login automatico atraves do delimitador '@'.
*  1.0: O inicio, estrutura simples (mais do que a atual :),
*        apenas 10 opcoes de shell eram permitidas. 
*
* Copyright (C) 2004 - 2006 Marcelo Leal.
*/

/*
*    ===========================================================
*    === Algumas explicacoes sobre as estruturas utilizadas ===
*    ===========================================================
*
*        ASHELLS: array de shells para o servidor em questao.
*        ADESCR: array contendo a descricao para cada shell.
*        PSHELLS: ponteiro para o shell corrente.
*        PDESCR: ponteiro para a descricao corrente.
*        Exemplo de sintaxe para o atributo LDAP:
*        shell: linpro101mtz=/bin/bash Shell de uso geral.
*
*    ===========================================================
*/

#define _GNU_SOURCE
#define TAM64 64
#define TAM128 128
#define TAM256 256
#define TAM512 512
#define BASHLOGIN "/bin/bash"
#include <stdio.h>
#include <stdlib.h>
#include <signal.h>
#include <errno.h>
#include <string.h>
#include <unistd.h>
#include <regex.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <pwd.h>
#include <syslog.h>
#include <lber.h>
#include <ldap.h>
#include <menu.h>
#include <curses.h>
#include <glib.h>

#define ARRAY_SIZE(tam) (sizeof(tam) / sizeof(tam[0]))
#define SHELL_WRAPPER_CONFIG_FILE  "/etc/shell_wrapper.conf"

#define _CONF_INTEGER	1
#define _CONF_STRING	2
#define CONF_INTEGER(x, y)		{ x, &y, _CONF_INTEGER, 0 }
#define CONF_STRING(x, y)		{ x, y, _CONF_STRING, sizeof(y) }
#define CONF_NULL			{ NULL, NULL, 0, 0 }
#define CONF_STRPTR_INTEGER(w, z)	CONF_INTEGER(#z, w->z)
#define CONF_STRPTR_STRING(w, z)	CONF_STRING(#z, w->z)

// Atributo shell recebido do LDAP
struct carga_ldap {
    char servidor[TAM512];
    char target[TAM512];
    char descricao[TAM64];
};
struct ldap_conf {
    char ldap_host[TAM64];
    char ldap_basedn[TAM64];
    char ldap_attr[TAM64];
};
struct ldap_readconf_parm {
    char *conf_parmname;
    void *conf_variable;
    int conf_type;
    int conf_maxlength;
};
struct sw_options {
    const char *config_file;
    int debug;
};

extern int h_errno;
static int valida_sintaxe (char *str, struct carga_ldap *pcarga);
static int ldap_readconf(const char *fname, struct ldap_conf *conf); 
static int sw_getopt(struct sw_options *opts, int argc, const char **argv); 
static int sw_configure( struct sw_options *opts, struct ldap_conf *conf, int argc, const char **argv); 
void print_in_middle(WINDOW *win, int starty, int startx, int width, char *string, chtype color);
void exec_binario (char *binario, const char **params, int AUTO);
void func(char *name);

int main (int argc, const char **argv) {
 struct passwd *pwd;
 struct sw_options opts;
 struct ldap_conf conf, *pconf = &conf;
 struct carga_ldap carga, *pcarga = &carga;
 pid_t mypid;
 LDAP *ld;
 BerElement *ber;
 LDAPMessage *ldap_message_set, *entry;
 char **ashells, **adescr, **vals, *attr, *cptmp, *payload;
 gchar *filter;
 char myhostname[TAM256], filtro1[TAM256], filtro2[TAM256]="todos_os_servidores=", *attributes[3] = {"dn", "shell", NULL};
 int TAMANHO=0, i, n=0, automatico, escolha, cod_retorno; 
 MENU *menu_principal;
 ITEM **itens_menu;
 WINDOW *menu_principal_win;

 // the game... :))
 mypid = getpid();
 
 cod_retorno = sw_configure(&opts, &conf, argc, argv);
 if (cod_retorno != 0) exit(EXIT_FAILURE);

 if (gethostname(myhostname, 255) != 0) {
  syslog(LOG_ERR, "[%d] ERRO Nao consegui determinar meu hostname.", mypid);
  exit(EXIT_FAILURE);
 } 
 strcpy (filtro1, myhostname);
 strcat (filtro1, "=");
 // Username...
 pwd=getpwuid(getuid());
 if (pwd == NULL) {
  syslog(LOG_ERR, "[%d] ERRO Usuario: [%s] desconhecido.", mypid, pwd->pw_name);
  exit(EXIT_FAILURE);
 }

 filter = g_strconcat ("(&(uid=", pwd->pw_name, ")", " (|(", pconf->ldap_attr, "=todos_os_servidores=*) (", pconf->ldap_attr, "=", myhostname, "=*", ")))", NULL);

 if ((ld = ldap_init (pconf->ldap_host, LDAP_PORT)) == NULL ) {
  syslog(LOG_ERR, "[%d] ERRO em ldap_init - LDAP Server: [%s]", mypid, pconf->ldap_host);
  exit(EXIT_FAILURE);  
 }

 if (ldap_search_s (ld, pconf->ldap_basedn, LDAP_SCOPE_SUBTREE, filter, attributes, 0, &ldap_message_set) != LDAP_SUCCESS) {
  syslog(LOG_ERR, "[%d] ERRO em ldap_search_s - BASE DN: [%s], FILTRO: [%s]", mypid, pconf->ldap_basedn, filter);
  exit(EXIT_FAILURE);  
 } 
 syslog(LOG_INFO, "[%d] FILTRO: [%s].", mypid, filter);
 g_free(filter);

 entry = ldap_first_entry(ld, ldap_message_set);
 if ( entry == NULL ) {
  syslog(LOG_INFO, "[%d] Usuario: [%s] sem shell configurado para este host.", mypid, pwd->pw_name);
  exit(EXIT_FAILURE);
 }
 // Quantos opcoes de shell temos para este servidor???
 for ( entry = ldap_first_entry(ld, ldap_message_set); entry != NULL; entry = ldap_next_entry(ld, entry)) { 
  attr = ldap_first_attribute (ld, entry, &ber);
  for (attr = ldap_first_attribute (ld, entry, &ber); attr != NULL; attr = ldap_next_attribute (ld, entry, ber)) {
   if ((vals = ldap_get_values(ld, entry, attr)) != NULL) {
    for (i = 0; vals[i] != NULL; i++) {
     if ((strncmp (vals[i], filtro1, strlen(filtro1)))==0 || ((strncmp (vals[i], filtro2, strlen(filtro2)))==0)) {
      TAMANHO++;
     }
    }
   }
  }
 }

 // Criacao do array (descricao e shell), e dos itens do Menu. 
 ashells = (char **) calloc(TAMANHO, sizeof(char *));
 adescr = (char **) calloc(TAMANHO, sizeof(char *));
 itens_menu = (ITEM **)calloc(TAMANHO, sizeof(ITEM *));
 if ((ashells == NULL) || (itens_menu == NULL) || (adescr == NULL)) {
  syslog(LOG_ERR, "[%d] ERRO na alocacao de memoria para composicao do menu de shells.", mypid);
  exit(EXIT_FAILURE);
 }
 char **pshells = ashells;
 char **pdescr = adescr;

 entry = ldap_first_entry(ld, ldap_message_set);
 if ( entry == NULL ) {
  syslog(LOG_INFO, "[%d] usuario: [%s] sem shell configurado para este host.", mypid, pwd->pw_name);
  exit(EXIT_FAILURE);
 }
 for ( entry = ldap_first_entry(ld, ldap_message_set); entry != NULL; entry = ldap_next_entry(ld, entry)) { 
  attr = ldap_first_attribute (ld, entry, &ber);
  for (attr = ldap_first_attribute (ld, entry, &ber); attr != NULL; attr = ldap_next_attribute (ld, entry, ber)) {
   if ((vals = ldap_get_values(ld, entry, attr)) != NULL) {
    for (i = 0; vals[i] != NULL; i++) {
     // Iremos avaliar apenas os atributos shell que interessam.
     if ((strncmp (vals[i], filtro1, strlen(filtro1)))==0 || ((strncmp (vals[i], filtro2, strlen(filtro2)))==0)) { 
      syslog(LOG_WARNING, "[%d] ITEM...[%d] [%s]", mypid, n, vals[i]);
      cptmp = strdupa (vals[i]);
      if (valida_sintaxe(cptmp, pcarga)) {
       syslog(LOG_ERR, "[%d] ITEM...[%d] [%s] ERRO de sintaxe. O shell nao esta separado de sua descricao.", mypid, n, vals[i]);
       strcpy(pcarga->descricao, "*** Opcao invalida, contate SAS/Unix ***");
       strcpy(pcarga->target, "/bin/false");
      } 
      asprintf (pdescr, pcarga->descricao);
      asprintf (pshells, pcarga->target);
      // Execucao direta...
      if (strstr(pcarga->descricao, "@") != NULL) {
       syslog(LOG_INFO, "[%d] Login AUTOMATICO para user [%s] neste servidor. SHELL [%s]", mypid, pwd->pw_name, pcarga->target);
       payload = strdupa (pcarga->target);
       automatico = 1;
       exec_binario(payload, argv, automatico);
       exit(EXIT_SUCCESS);
      }
      itens_menu[n] = new_item((char *)*pdescr, (char *)*pdescr);
      pshells++; pdescr++; n++;
     }
    }
    ldap_value_free(vals);
   }
   // A linha a seguir foi comentada, por estar causando um segmentation fault.
   // ldap_memfree(attr);
  }
  itens_menu[TAMANHO] = (ITEM *)NULL;
  if (ber != NULL) {
   ber_free(ber,0);
  }
 }
 ldap_msgfree(ldap_message_set);
 
 // Formatacao da tela... inicio do ncurses. 
 // Thanks to: <Pradeep Padala> pelos exemplos... :))
 initscr();
 start_color();
 cbreak();
 noecho();
 keypad(stdscr, TRUE);
 init_pair(1, COLOR_RED, COLOR_BLACK);
 init_pair(2, COLOR_CYAN, COLOR_BLACK);
 init_pair(3, COLOR_MAGENTA, COLOR_BLACK);
 
 // Criacao do Menu principal 
 menu_principal = new_menu((ITEM **)itens_menu);
 // Nao mostrar a descricao de cada opcao...
 menu_opts_off(menu_principal, O_SHOWDESC);
 // Criacao da janela a ser associada com o menu 
 menu_principal_win = newwin(14, 70, 4, 4);
 keypad(menu_principal_win, TRUE);
 // Janela principal e sub menu.
 set_menu_win(menu_principal, menu_principal_win);
 set_menu_sub(menu_principal, derwin(menu_principal_win, 11, 68, 3, 1));
 set_menu_format(menu_principal, 10, 1);
 // Cursor 
 set_menu_mark(menu_principal, " -> ");
 // Borda e titulo do menu 
 box(menu_principal_win, 0, 0);
 print_in_middle(menu_principal_win, 1, 15, 40, "Menu de Operaes", COLOR_PAIR(2));
 mvwaddch(menu_principal_win, 2, 0, ACS_LTEE);
 mvwhline(menu_principal_win, 2, 1, ACS_HLINE, 68);
 mvwaddch(menu_principal_win, 2, 69, ACS_RTEE);
 
 // Menu 
 post_menu(menu_principal);
 wrefresh(menu_principal_win);

 attron(COLOR_PAIR(2));
 mvprintw(LINES - 5, 26, "Pressione: <Ctrl C> para SAIR");
 mvprintw(LINES - 3, 35, "-MYCOMPANY-");
 mvprintw(LINES - 2, 10, "------------------------------------------------------------");
 attroff(COLOR_PAIR(2));
 refresh();

 while((escolha = wgetch(menu_principal_win)) != KEY_F(1)) {       
  switch(escolha) {
   case KEY_DOWN:
    menu_driver(menu_principal, REQ_DOWN_ITEM);
    break;
   case KEY_UP:
    menu_driver(menu_principal, REQ_UP_ITEM);
    break;
   case 10: { /* Enter */
    ITEM *cur;
    cur = current_item(menu_principal);
    syslog(LOG_INFO, "[%d] Aplicacao escolhida para sessao: [%s] - [%s]", mypid, ashells[cur->index], adescr[cur->index]);
    payload = strdupa (ashells[cur->index]);
    unpost_menu(menu_principal);
    for(i = 0; i < TAMANHO; ++i) free_item(itens_menu[i]);
    free_menu(menu_principal);
    endwin();
    automatico = 0;
    exec_binario(payload, argv, automatico);
    exit(EXIT_SUCCESS);
   }
  }
 }
 exit(EXIT_SUCCESS);
}

void print_in_middle(WINDOW *win, int starty, int startx, int width, char *string, chtype color) {
   int length, x, y;
   float temp;
   
   if(win == NULL)
   win = stdscr;
   getyx(win, y, x);
   if(startx != 0)
   x = startx;
   if(starty != 0)
   y = starty;
   if(width == 0)
   width = 80;

   length = strlen(string);
   temp = (width - length)/ 2;
   x = startx + (int)temp;
   wattron(win, color);
   mvwprintw(win, y, x, "%s", string);
   wattroff(win, color);
   refresh();
}

// Ler configuracoes..
static int ldap_readconf(const char *fname, struct ldap_conf *conf) {
    FILE *f;
    struct ldap_readconf_parm ldap_parms[] = {
	CONF_STRPTR_STRING(conf, ldap_host),
	CONF_STRPTR_STRING(conf, ldap_basedn),
	CONF_STRPTR_STRING(conf, ldap_attr),
	CONF_NULL,
    };
    struct ldap_readconf_parm *p;
    char buffer[TAM512];
    char *vname, *end, *value, *errorptr;
    int intvalue, ln, length;
    pid_t mypid;

    mypid = getpid();
    if (!(f = fopen(fname, "r"))) {
     syslog(LOG_ERR, "[%d] ERRO acessando arquivo de configuracao: [%s], trabalhando com valores padroes. host: [%s], basedn: [%s].", mypid, fname, conf->ldap_host, conf->ldap_basedn);
     return 0; 
    }
    ln = 0;
    while (fgets(buffer, sizeof(buffer), f)) {
	ln++;
	length = strlen(buffer);
	if (buffer[length - 1] != '\n') {
	    syslog(LOG_WARNING, "[%d] linha muito longa em [%s]:[%d], ignorada.", mypid, fname, ln);
	    while (fgets(buffer, sizeof(buffer), f)) {
		if (buffer[strlen(buffer) - 1] == '\n') break;
	    }
	    continue;
	}
	buffer[length - 1] = '\0';
	for (vname = buffer; *vname == ' ' || *vname == '\t'; vname++);
	if (*vname == '#' || *vname == '\0') continue;
	if (!(value = strchr(vname, '='))) {
	    syslog(LOG_WARNING, "[%d] erro de sintaxe em [%s]:[%d]", mypid, fname, ln);
	    continue;
	}
	for (end = value - 1; *end == ' ' || *end == '\t'; end--);
	end[1] = '\0';
	for (p = ldap_parms; p->conf_parmname; p++) {
	    if (strcmp(p->conf_parmname, vname) == 0) break;
	}
	if (!p->conf_parmname) {
	    syslog(LOG_WARNING, "[%d] variavel desconhecida [%s] na linha [%s]:[%d]", mypid, vname, fname, ln);
	    continue;
	}
	for (value++; *value == ' ' || *value == '\t'; value++);
	for (end = buffer + length - 2; *end == ' ' || *end == '\t'; end--);
	end++; *end = '\0';
	switch (p->conf_type) {
	case _CONF_INTEGER:
	    intvalue = strtol(value, &errorptr, 0);
	    if (*errorptr != '\0') {
		syslog(LOG_WARNING, "[%d] valor invalido para a variavel [%s] em [%s]:[%d]", mypid, vname, fname, ln);
		continue;
	    }
	    *((int *) p->conf_variable) = intvalue;
	    break;
	case _CONF_STRING:
	    if ((length = end - value) >= p->conf_maxlength) {
		syslog(LOG_WARNING, "[%d] valor muito grande para variavel [%s] em [%s]:[%d]", mypid, vname, fname, ln);
		continue;
	    }
	    memcpy((char *) p->conf_variable, value, length + 1);
	    break;
	default:
	    /* ignore */
	    break;
	}
    }
    fclose(f);
  return 0;
}

static int sw_getopt(struct sw_options *opts, int argc, const char **argv) {
 int i;
 pid_t mypid;

 mypid = getpid();
 opts->config_file = SHELL_WRAPPER_CONFIG_FILE;
 opts->debug = 0;
 // Vamos imprimir os argumentos, caso existam.
 for (i = 1; i < argc; i++) {
  syslog(LOG_WARNING, "[%d] ARGUMENTO [%i]: [%s]", mypid, i, argv[i]);
 }
 return 0;
}

static int sw_configure( struct sw_options *opts, struct ldap_conf *conf, int argc, const char **argv) {
 int cod_retorno;

 cod_retorno = sw_getopt(opts, argc, argv);
 if (cod_retorno != 0) return cod_retorno;

 /* inicializa as opcoes ldap */
 strcpy(conf->ldap_host, "ldap.yourcompany");
 strcpy(conf->ldap_basedn, "o=yourcompany,c=cia");
 strcpy(conf->ldap_attr, "shell");

 /* Nossas opcoes do arquivo de configuracao */
 cod_retorno = ldap_readconf(opts->config_file, conf);
 return cod_retorno;
}

int valida_sintaxe (char *str, struct carga_ldap *pcarga) {
 char *p;

 if (( p = strtok (str,"=")) == NULL ) {
  return 1;
 } else {
  if (strlen(p) > TAM64) {
   return 1;
  } else {
   strcpy (pcarga->servidor, p);
  }
 }
 if (( p = strtok (NULL, " ")) == NULL ) {
  return 1;
 } else {
  if (strlen(p) > TAM64) {
   return 1;
  } else {
   strcpy (pcarga->target, p);
  }
 }
 if (( p = strtok (NULL, "")) == NULL ) {
  return 1;
 } else {
  if (strlen(p) > TAM64) {
   return 1;
  } else {
   strcpy (pcarga->descricao, p);
  }
 }
 return 0;
}

void exec_binario (char *binario, const char **param, int AUTO) {

 if ((strcmp(binario, BASHLOGIN) == 0) && (AUTO)) {
  if ( param == NULL ) {
   execl (binario, binario, "--login", NULL, (char *) 0);
  } else {
   execv (binario, (char **) param);
  }
 }

 if ((strcmp(binario, BASHLOGIN) == 0) && (AUTO == 0)) { 
  execl (binario, binario, "--login", NULL, (char *) 0); 
 }

 if ((AUTO) && (strcmp(binario, BASHLOGIN) != 0)) {
   execv (binario, (char **) param);
 } else {
  execl (binario, binario, NULL, NULL, (char *) 0); 
 }
}
